﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.IO;
using System.Threading;

namespace IRcmd
{
    public class IRPacket
    {
        public int header_h, header_l;
        public int data1_h, data1_l;
        public int data0_h, data0_l;
        public int trailer_h, trailer_l;
        public int len;
        public byte[] data;

        public IRPacket()
        {
            header_h = header_l =
            data1_h = data1_l =
            data0_h = data0_l =
            trailer_h = trailer_l =
            len = 0;
            data = null;
        }

        public void alloc_data(int bits)
        {
            len = bits;
            data = new byte[bytes()];
        }

        public int bytes()
        {
            return (len - 1) / 8 + 1;
        }

        public byte checksum()
        {
            //byte cs = (byte)(header_h + header_l +
            byte cs = (byte)(
                (byte)(header_h / 256) + (byte)(header_h & 0xff) +
                (byte)(header_l / 256) + (byte)(header_l & 0xff) +
                data1_h + data1_l +
                data0_h + data0_l +
                trailer_h + trailer_l);

            cs += (byte)(len >> 8);
            cs += (byte)(len & 0xff);

            for (int i = 0; i < bytes(); i++)
            {
                cs += data[i];
            }
            return cs;
        }

        override public String ToString()
        {
            String ret = "";
            ret += "HH:" + header_h + " \n";
            ret += "HL:" + header_l + " \n";
            ret += "D1H:" + data1_h + " \n";
            ret += "D1L:" + data1_l + " \n";
            ret += "D0H:" + data0_h + " \n";
            ret += "D0L:" + data0_l + " \n";
            ret += "TH:" + trailer_h + " \n";
            ret += "TL:" + trailer_l + " \n";
            ret += "LEN:" + len + " \n";
            for (int i = 0; i < bytes(); i++)
            {
                ret += data[i].ToString("X2") + " ";
            }
            return ret;
        }

        public void save(string fname)
        {
            StreamWriter sw = new StreamWriter(fname);
            save(sw);
            sw.Close();
        }

        public void save(TextWriter w)
        {
            w.Write("IPPACKET001\n");
            w.Write(ToString());
            w.Write("\n");
        }

        static public void save(string fname, IRPacket[] ps)
        {
            StreamWriter sw = new StreamWriter(fname);
            foreach(IRPacket p in ps) p.save(sw);
            sw.Close();
        }

        static private IRPacket load_one(TextReader sr)
        {
            IRPacket ret = new IRPacket();

            string line;
            int n = 0;
            string[] ss;
            for (n = 0; n < 11; n++)
            {
                line = sr.ReadLine();
                if (line == null) return null;
                switch (n)
                {
                    case 0:
                        if (line != "IPPACKET001")
                        {
                            sr.Close();
                            throw new Exception("Invalid format");
                        }
                        break;
                    case 1:
                        ss = line.Split(':', ' ');
                        ret.header_h = int.Parse(ss[1]);
                        break;
                    case 2:
                        ss = line.Split(':', ' ');
                        ret.header_l = int.Parse(ss[1]);
                        break;
                    case 3:
                        ss = line.Split(':', ' ');
                        ret.data1_h = int.Parse(ss[1]);
                        break;
                    case 4:
                        ss = line.Split(':', ' ');
                        ret.data1_l = int.Parse(ss[1]);
                        break;
                    case 5:
                        ss = line.Split(':', ' ');
                        ret.data0_h = int.Parse(ss[1]);
                        break;
                    case 6:
                        ss = line.Split(':', ' ');
                        ret.data0_l = int.Parse(ss[1]);
                        break;
                    case 7:
                        ss = line.Split(':', ' ');
                        ret.trailer_h = int.Parse(ss[1]);
                        break;
                    case 8:
                        ss = line.Split(':', ' ');
                        ret.trailer_l = int.Parse(ss[1]);
                        break;
                    case 9:
                        ss = line.Split(':', ' ');
                        ret.len = int.Parse(ss[1]);
                        ret.alloc_data(ret.len);
                        break;
                    case 10:
                        ss = line.Split(' ');
                        for (int i = 0; i < ret.bytes(); i++)
                        {
                            ret.data[i] = byte.Parse(ss[i], System.Globalization.NumberStyles.HexNumber);
                        }
                        break;
                }
            }
            return ret;
        }

        static public IRPacket[] load(string fname)
        {
            List<IRPacket> ret = new List<IRPacket>();
            StreamReader sr = new StreamReader(fname);

            IRPacket p;
            while ((p = load_one(sr)) != null) ret.Add(p);
            sr.Close();

            if (ret.Count == 0) return null;

            return ret.ToArray();
        }
    }


    public class IRShield
    {
        protected SerialPort port;
        public IRPacket pkt;

        public IRShield(SerialPort p)
        {
            port = p;
            pkt = new IRPacket();
        }

        public int send(IRPacket p)
        {
            return send(p, false);
        }

        public int send(IRPacket p, bool chk_ack)
        {
            port.ReadTimeout = 1000;

            List<byte> b = new List<byte>();

            if (p.len <= 0 || p.data == null) return 0;
            b.Add((byte)0xff);  // sync
            
            b.Add((byte)(p.header_h >> 8));
            b.Add((byte)(p.header_h & 0xff));
            b.Add((byte)(p.header_l >> 8));
            b.Add((byte)(p.header_l & 0xff));

            b.Add((byte)p.data1_h);
            b.Add((byte)p.data1_l);
            b.Add((byte)p.data0_h);
            b.Add((byte)p.data0_l);
            b.Add((byte)p.trailer_h);
            b.Add((byte)p.trailer_l);

            b.Add((byte)(p.len >> 8));  // data length
            b.Add((byte)(p.len & 0xff));

            for (int i = 0; i < p.bytes(); i++)
            {
                b.Add(p.data[i]);
            }

            byte cs = p.checksum();
            b.Add(cs);

            try
            {
                if (chk_ack) read_data(); fifo.Clear();

                port.Write(b.ToArray(), 0, b.Count);

                if (chk_ack && read_byte() != cs)
                {
                    throw new Exception("Check sum does not match.");
                }
            }
            catch (Exception e)
            {
                Console.Error.WriteLine(e.ToString());
                return -1;
            }
            return p.len;
        }

        public IRPacket receive()
        {
            return receive(500);
        }

        public IRPacket receive(int timuout)
        {
            port.ReadTimeout = timuout;    // timeout (msec)

            if (port.BytesToRead == 0 && fifo.Count == 0) return null;
            if (read_byte() == 0xff)
            {   // sync
                IRPacket ret = new IRPacket();
                if (!recieve_length(ret))
                {
                    throw new Exception("palse length error");
                }
                ret.len = read_byte() * 256 + read_byte();
                if (ret.len > 400 || ret.len <= 0)
                {
                    throw new Exception("out of length");
                }
                ret.data = new byte[ret.bytes()];
                for (int i = 0; i < ret.bytes(); i++)
                {
                    ret.data[i] = read_byte();
                }
                // checksum
                byte cs = read_byte();
                if (cs != ret.checksum())
                {
                    throw new Exception("checksum error");
                }
                return ret;
            }
            else
            {
                return null;
            }
        }

        private bool recieve_length(IRPacket p)
        {
            /*
            p.header_h = read_byte();
            if (p.header_h == 0) return false;
            p.header_l = read_byte();
            if (p.header_l == 0) return false;
            */
            p.header_h = read_byte() * 256 + read_byte();
            if (p.header_h == 0) return false;
            p.header_l = read_byte() * 256 + read_byte();
            if (p.header_l == 0) return false;
            p.data1_h = read_byte();
            if (p.data1_h == 0) return false;
            p.data1_l = read_byte();
            if (p.data1_l == 0) return false;
            p.data0_h = read_byte();
            if (p.data0_h == 0) return false;
            p.data0_l = read_byte();
            if (p.data0_l == 0) return false;
            p.trailer_h = read_byte();
            if (p.trailer_h == 0) return false;
            p.trailer_l = read_byte();
            if (p.trailer_l != 255) return false;
            return true;
        }

        private Queue<byte> fifo = new Queue<byte>();

        private int read_data()
        {
            if (port.BytesToRead == 0) return 0;
            byte[] buf = new byte[port.BytesToRead];
            port.Read(buf, 0, buf.Length);

            foreach (byte b in buf)
                fifo.Enqueue(b);

            return buf.Length;
        }

        private byte read_byte()
        {
            if (fifo.Count == 0)
            {
                int n = 0;
                while (read_data() == 0)
                {
                    if (n++ > port.ReadTimeout)
                        throw new Exception("serial read timeout");
                    Thread.Sleep(1);
                }
            }
            return fifo.Dequeue();
        }
    }
}
